/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) 2016 OX Software GmbH
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/portal/portaltour', [
    'io.ox/office/tk/config',
    'io.ox/office/tk/utils/driveutils',
    'io.ox/office/baseframework/app/guidedtour',
    'io.ox/office/portal/portalutils',
    'gettext!io.ox/office/portal/main'
], function (Config, DriveUtils, GuidedTour, PortalUtils, gt) {

    'use strict';

    // the 'shown' configuration key for any of the portal tours
    var ANY_PORTAL_TOUR_SHOWN_KEY = 'portal/alltours/shown';

    // selector for any tool pane without hidden contents
    var TOOL_PANE_SELECTOR = '.tool-pane:not(.hidden-contents)';

    // bug 51700: do not show welcome tours if the standard tour package is not installed
    var TOURS_PACKAGE_INSTALLED = Config.getDebugUrlFlag('office:show-tours') || require.defined('io.ox/tours/main');

    // debug mode: whether to always show the long version of the tours
    var DEBUG_FULL_TOURS = Config.getDebugUrlFlag('office:full-tours');

    // class PortalTour =======================================================

    /**
     * Base class for all guided tours for the Portal applications.
     *
     * @constructor
     *
     * @extends GuidedTour
     *
     * @param {String} appBaseName
     *  The base name of the application represented by this tour.
     */
    var PortalTour = GuidedTour.extend(function (appBaseName, initOptions) {

        // self reference
        var self = this;

        // type identifier of the Portal application
        var PORTAL_APP_TYPE = PortalUtils.getPortalModuleName(appBaseName);

        // type identifier of the associated editor application
        var EDITOR_APP_TYPE = PortalUtils.getEditorModuleName(appBaseName);

        // CSS selector for the 'My Files/Documents' folder in Drive
        var DOCUMENTRS_FOLDER_SELECTOR = '.folder-tree .folder[data-id="' + PortalUtils.getNewLaunchOptions().target_folder_id + '"] .folder-node';

        // the editor application (may exist but be inactive, e.g. when using Back button)
        var editApp = null;

        // the launch options used to show the existing editor application
        var editLaunchOptions = null;

        // whether to show the short version of the tour
        var shortTour = !DEBUG_FULL_TOURS && (initOptions && initOptions.auto && PortalTour.isAnyPortalTourShown());

        // base constructor ---------------------------------------------------

        GuidedTour.call(this);

        // constants ----------------------------------------------------------

        // selector for the root window node of the editor application
        this.EDITOR_WINDOW_SELECTOR = '.window-container[data-app-name="' + EDITOR_APP_TYPE + '"][data-doc-loaded="true"]:visible';

        // selector for the top pane of the editor application
        this.EDITOR_TOP_PANE_SELECTOR = this.EDITOR_WINDOW_SELECTOR + ' ' + TOOL_PANE_SELECTOR + '.top-pane';

        // selector for the main tool pane of the editor application
        this.EDITOR_TOOL_PANE_SELECTOR = this.EDITOR_WINDOW_SELECTOR + ' ' + TOOL_PANE_SELECTOR + '.edit-pane';

        // selector for the application pane of the editor application
        this.EDITOR_APP_PANE_SELECTOR = this.EDITOR_WINDOW_SELECTOR + ' .app-pane';

        // public methods -----------------------------------------------------

        /**
         * Returns whether to show the short version of this tour.
         *
         * @returns {Boolean}
         *  Whether to show the short version of this tour.
         */
        this.isShortTour = function () {
            return shortTour;
        };

        // specific portal tour steps -----------------------------------------

        /**
         * Immediately launches the Drive application.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the Drive application has been
         *  launched.
         */
        this.launchDriveApp = function () {
            return ox.launch('io.ox/files/main');
        };

        /**
         * Immediately launches the Portal application of this tour.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the Portal application has
         *  been launched.
         */
        this.launchPortalApp = function () {
            return ox.launch(PORTAL_APP_TYPE + '/main');
        };

        /**
         * Immediately launches an existing editor application, if a blank
         * document has been created already.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the editor application has
         *  been launched.
         */
        this.launchEditorApp = function () {
            return editApp ? ox.launch(EDITOR_APP_TYPE + '/main', editLaunchOptions) : $.Deferred().reject();
        };

        /**
         * Defines the leading welcome step of the portal tour, referring to
         * the launcher tab of the portal application.
         *
         * @attention
         *  The created tour step will neither contain a title nor any content.
         *
         * @returns {Step}
         *  The new tour step. The instance has been extended with the mix-in
         *  class StepMixin defined in the GuidedTour module.
         */
        this.welcomeStep = function () {
            return this.step()
                .referTo('#io-ox-topbar .launcher[data-app-name="' + PORTAL_APP_TYPE + '"]', { position: 'bottom', hotspot: true });
        };

        /**
         * Defines the tour step referring to the templates area in the portal
         * application window.
         *
         * @attention
         *  The created tour step will neither contain a title nor any content.
         *
         * @returns {Step}
         *  The new tour step. The instance has been extended with the mix-in
         *  class StepMixin defined in the GuidedTour module.
         */
        this.templatesAreaStep = function () {
            return this.step()
                .hotspot('.office-portal-templates h2')
                .referTo('.office-portal-templates .template-item:first-child', { position: 'bottom' });
        };

        /**
         * Defines the tour step referring to the 'Blank document' button of
         * the portal application.
         *
         * @attention
         *  The created tour step will neither contain a title nor any content.
         *
         * @returns {Step}
         *  The new tour step. The instance has been extended with the mix-in
         *  class StepMixin defined in the GuidedTour module.
         */
        this.blankDocumentStep = function () {
            return this.step()
                // wait for the element in case this step is shown via Back button from next step
                .waitForAndReferTo('.office-portal-templates .template-item.blank-doc', { position: 'bottom', hotspot: '.preview-box' });
        };

        /**
         * Defines the tour step that creates a new blank document, activates
         * the editor application, waits for launching, and refers to the tool
         * pane of the application window. The launch step will already contain
         * a title and a description text.
         *
         * @returns {Step}
         *  The new tour step. The instance has been extended with the mix-in
         *  class StepMixin defined in the GuidedTour module.
         */
        this.launchDocumentStep = function () {
            return this.step()
                .title(gt.pgettext('tour', 'Toolbar'))
                .content(gt.pgettext('tour', 'At the top of the screen you will find the toolbar which shows all the common functions.'))
                // register the application launcher, create a new blank document
                .navigateTo(EDITOR_APP_TYPE + '/main', PortalUtils.getNewLaunchOptions())
                // wait for the application window node, and the 'data-doc-loaded' marker that will be set after import
                .waitFor(this.EDITOR_TOP_PANE_SELECTOR, 10)
                // highlight the entire edit tool pane (bug 44195: use first embedded tool bar element to prevent positioning problems with full-width pane)
                .waitForAndReferTo(this.EDITOR_TOOL_PANE_SELECTOR + ' .toolbar:visible:eq(0)', { timeout: 10, position: 'bottom', hotspot: true })
                // once the document has been created, change this step so that it navigates to the existing document
                .once('show', function () {
                    editApp = this.getApp();
                    editLaunchOptions = { action: 'load', file: editApp.getFileDescriptor() };
                    this.navigateTo(EDITOR_APP_TYPE + '/main', editLaunchOptions);
                })
                // launch the source application when leaving this step backwards
                .on('back', this.launchPortalApp.bind(this));
        };

        /**
         * Defines the tour step that inserts some contents in the new
         * document created by the launch step, and refers to the application
         * status label shown in the top pane. The auto-save step will already
         * contain a title and a description text.
         *
         * @param {Function} callback
         *  A callback function that must insert some contents into the new
         *  document, in order to show the 'Saving changes' label. Will only be
         *  called once when reaching this step the first time.
         *
         * @returns {Step}
         *  The new tour step. The instance has been extended with the mix-in
         *  class StepMixin defined in the GuidedTour module.
         */
        this.autoSaveStep = function (callback) {
            this.step()
                .title(gt.pgettext('tour', 'Autosaving changes'))
                .content(gt.pgettext('tour', 'Changes are saved automatically whenever you work on your document.'))
                // invoke the callback function that will insert some contents into the document
                .once('before:show', callback)
                // wait for the application window node, and the 'data-doc-loaded' marker that will be set after import
                .waitFor(this.EDITOR_TOP_PANE_SELECTOR + ' .app-status-label .caption i')
                // highlight the 'All changes saved' status label
                .referTo(this.EDITOR_TOP_PANE_SELECTOR + ' .app-status-label', { position: 'bottom', hotspot: '.caption i' });
        };

        /**
         * Defines the tour step that refers to the tab button of the 'File'
         * toolbar. The step will already contain a title and a description
         * text.
         *
         * @returns {Step}
         *  The new tour step. The instance has been extended with the mix-in
         *  class StepMixin defined in the GuidedTour module.
         */
        this.fileTabStep = function () {
            return this.step()
                .title(gt.pgettext('tour', 'File tab'))
                .content(gt.pgettext('tour', 'The file tab offers you various ways to share or save your document.'))
                .on('before:show', function () { this.triggerOptionButton(self.EDITOR_TOP_PANE_SELECTOR, 'view/toolbars/tab', 'file'); })
                .hotspot(GuidedTour.getOptionButtonSelector(this.EDITOR_TOP_PANE_SELECTOR, 'view/toolbars/tab', 'file'))
                // do not cover the edit tool pane with the pop-up box
                .referTo(this.EDITOR_TOOL_PANE_SELECTOR, { position: 'bottom' });
        };

        /**
         * Defines the tour step that opens and refers to the 'Save In Drive'
         * drop-down menu. The save-as step will already contain a title and a
         * description text.
         *
         * @returns {Step}
         *  The new tour step. The instance has been extended with the mix-in
         *  class StepMixin defined in the GuidedTour module.
         */
        this.saveAsStep = function () {
            return this.step()
                .title(gt.pgettext('tour', 'Saving options'))
                .content(gt.pgettext('tour', 'If you would like to store your document in a specific folder, just click on "Save as". You can also click on "Save as template" to create your own templates.'))
                .content($('<br>'))
                .content(gt.pgettext('tour', 'You can also click on "Save as template" to create your own templates.'))
                // open and highlight the 'Save As' drop-down menu
                .showDropDownMenu(this.EDITOR_TOOL_PANE_SELECTOR, 'view/saveas/menu')
                .waitForAndReferToPopup({ position: 'right', hotspot: true });
        };

        /**
         * Defines the tour step that refers to the 'Send As Mail' button. The
         * mail step will already contain a title and a description text.
         *
         * @param {Obejct} [options]
         *  Optional parameters:
         *  @param {Boolean} [options.asDropDown=false]
         *      If set to true, the "Send as mail" control is a drop-down menu
         *      button, instead of a simple push button.
         *
         * @returns {Step}
         *  The new tour step. The instance has been extended with the mix-in
         *  class StepMixin defined in the GuidedTour module.
         */
        this.sendAsMailStep = function (options) {
            var groupKey = (options && options.asDropDown) ? 'document/sendmail/menu' : 'document/sendmail';
            return this.step()
                .title(gt.pgettext('tour', 'Sending your document'))
                .content(gt.pgettext('tour', 'Sending your document by email is easy. Just click on the mail icon and you are ready to go!'))
                .referToGroup(this.EDITOR_TOOL_PANE_SELECTOR, groupKey, { position: 'bottom', hotspot: '.caption i' });
        };

        /**
         * Defines the tour step that refers to the 'Close' button in the top
         * bar, without actually closing the document. The step will already
         * contain a title and a description text.
         *
         * @returns {Step}
         *  The new tour step. The instance has been extended with the mix-in
         *  class StepMixin defined in the GuidedTour module.
         */
        this.closeButtonStep = function () {
            return this.step()
                .title(gt.pgettext('tour', 'Closing document'))
                .content(gt.pgettext('tour', 'The document can be closed now. As all changes are stored automatically, you do not need to save documents manually.'))
                // wait for the element in case this step is shown via Back button from next step
                .waitForAndReferToGroup(this.EDITOR_TOP_PANE_SELECTOR, 'app/quit', { position: 'bottom', hotspot: '.caption i' });
        };

        /**
         * Defines the tour step that launches the 'Drive' application, and
         * refers to the 'Documents' folder. The step will already contain a
         * title and a description text.
         *
         * @returns {Step}
         *  The new tour step. The instance has been extended with the mix-in
         *  class StepMixin defined in the GuidedTour module.
         */
        this.documentsFolderStep = function () {
            return this.step()
                .title(gt.pgettext('tour', 'Documents folder'))
                .content(gt.pgettext('tour', 'You can find your document in your "Documents" folder.'))
                // launch the Drive application
                .navigateTo('io.ox/files/main')
                // wait for and highlight the 'Documents' folder (TODO: ensure visible folder side-pane)
                .waitForAndReferTo('.window-container[data-app-name="io.ox/files"] ' + DOCUMENTRS_FOLDER_SELECTOR, { position: 'bottom', hotspot: '.folder-label' })
                // launch the editor application when leaving this step backwards
                .on('back', this.launchEditorApp.bind(this));
        };

        /**
         * Defines the tour step that opens the settings drop-down menu, and
         * refers to the 'Restart This Tour' menu entry. The step will already
         * contain a title and a description text.
         *
         * @returns {Step}
         *  The new tour step. The instance has been extended with the mix-in
         *  class StepMixin defined in the GuidedTour module.
         */
        this.restartTourStep = function () {
            return this.step()
                .title(gt.pgettext('tour', 'Restart Guided Tour'))
                .content(gt.pgettext('tour', 'Hint: you can start guided tours, any time you need them, from the system menu.'))
                // show the Portal application again
                .navigateTo(PORTAL_APP_TYPE + '/main')
                // open and close the global drop-down menu
                .on('navigate before:hide', function () { $('#io-ox-topbar-dropdown-icon>a').dropdown('toggle'); })
                // wait for and highlight the 'Restart tour' menu entry
                .waitForAndReferTo('#topbar-settings-dropdown [data-action="guided-tour"]', { position: 'left', hotspot: true });
        };

        // initialization -----------------------------------------------------

        // set the 'shown' configuration flag for any of the portal welcome tours,
        // needed to distinguish between full tours and shortened tours
        this.on('start', function () {
            Config.set(ANY_PORTAL_TOUR_SHOWN_KEY, true);
        });

        // delete the temporary document in Drive when closing the tour
        this.on('stop', function () {

            // launch the portal application (the tour may have been canceled elsewhere)
            this.launchPortalApp();

            // close the editor application, and delete the document from Drive
            if (editApp) {
                var file = editApp.getFileDescriptor();
                editApp.quit('tour').done(function () {
                    var def = DriveUtils.purgeFile(file);
                    def.then(function () {
                        ox.trigger('refresh^');
                    });
                });
            }
        });

    }); // class PortalTour

    // static methods ---------------------------------------------------------

    /**
     * Returns whether any of the welcome tours for a Documents portal
     * application has already been shown.
     *
     * @returns {Boolean}
     *  Whether any of the welcome tours for a Documents portal application has
     *  already been shown.
     */
    PortalTour.isAnyPortalTourShown = function () {
        return Config.getFlag(ANY_PORTAL_TOUR_SHOWN_KEY);
    };

    /**
     * Registers a welcome tour for the specified Documents portal application.
     *
     * @param {string} appBaseName
     *  The base name of the portal application ('text', 'spreadsheet', etc.).
     */
    PortalTour.registerPortalTour = function (appBaseName) {
        if (!_.device('smartphone')) {
            var moduleName = PortalUtils.getPortalModuleName(appBaseName);
            var requirePath = moduleName + '/fulltour';
            var shownKey = PortalUtils.getPortalAppName(appBaseName) + '/fulltour/shown';
            GuidedTour.registerAppTour(moduleName, requirePath, shownKey);
        }
    };

    /**
     * Starts the welcome tour for the specified Documents portal application
     * once.
     *
     * @param {string} appBaseName
     *  The base name of the portal application ('text', 'spreadsheet', etc.).
     *
     * @returns {JQuery.Promise<PortalTour>}
     *  A promise that will be resolved with the running tour instance; or that
     *  will be rejected, if the tour has not been started (e.g. was already
     *  started before).
     */
    PortalTour.runPortalTour = function (appBaseName) {
        // start the welcome tour for the application once (and never in Selenium tests)
        // bug 51700: do not show any tours if the standard tour package is not installed
        if (!Config.AUTOTEST && TOURS_PACKAGE_INSTALLED) {
            var moduleName = PortalUtils.getPortalModuleName(appBaseName);
            return GuidedTour.runAppTour(moduleName, { once: true, auto: true });
        }
    };

    /**
     * Hides the menu entry "Show tour for this app" in the global settings
     * drop-down menu for the specified Documents portal application.
     *
     * @param {string} appBaseName
     *  The base name of the portal application ('text', 'spreadsheet', etc.).
     */
    PortalTour.updateSettingsDropDownMenu = function (appBaseName) {
        // workaround for bug 51700: hide entry in settings drop-down if the standard tour package is not installed
        if (!TOURS_PACKAGE_INSTALLED) {
            var moduleName = PortalUtils.getPortalModuleName(appBaseName);
            $('#topbar-settings-dropdown li[data-id="default/' + moduleName + '"]').remove();
        }
    };

    // exports ================================================================

    return PortalTour;

});
